/***************************************************************************/
// multiply lots of SU3 matrices and vectors:
// b[indx] = M[indx]*a[indx]

// define FORMAT_2COL in include file or in Makefile to enable 2-column SU(3)
// matrix storage
// change hardcoded VOLUME in include file to change problem size

// executable requires no command line arguments

#include "include_2col.h"

/**********************************************************************/
double dtime()
{
  // timing

    double tseconds = 0.0;
    struct timeval mytime;
    gettimeofday(&mytime,(struct timezone*)0);
    tseconds = (double)(mytime.tv_sec + mytime.tv_usec*1.0e-6);
    return( tseconds );
}


/**********************************************************************/
/************************************************************************/


int main(int argc, char *argv[] ) {

  //* multiply 3-component complex vectors by 3x3 complex matricies



  su3vec vecA;   // source vector
  su3vec vecB;   // result vector
  su3mat matM;   // SU(3) matrix


  int i,j;       // row, column indicies
  int rep;          // repeat index
  int num_reps=1000; // number of repeats to perform 

  int indx;         // index over "VOLUME"
  int tindex;       // thread or block index


  int hi, lo, offset;           // for parsing the long loop between threads
  double tstart, tstop, ttime;  // some timing info
  double gflops;                // number of Gflops performed           
  double gflop_per_sec;     // performance 
  double gbytes;              // data moved

  double bandwidth=0.0;       // data moved/sec 
  double intensity;           // flops/byte

  float total_data_size;      // total size of the problem (in + out)

  int numthreads;
  
  int ops_per_iter_1;        // in the main part of the calculation
  int bytes_per_iter_1;          // in the main part
  
  int ops_per_iter_2;       // in the reconstruction part of calculation
  int bytes_per_iter_2;          // in the reconstruction
  
  int num_blocks;                // default is number of threads
  int blocksize;                 // roughly  VOLUME/num_blocks



    // Start some threads to get the number of threads from 
    //OMP_NUM_THREADS environment variable
#pragma omp parallel
#pragma omp master
  numthreads = omp_get_num_threads();
  printf("Using %d threads\n",numthreads);


  printf("VOLUME= %d\n",VOLUME);
  // data size in MB
  total_data_size=(2*sizeof(su3vec)+sizeof(su3mat));
  printf("\ntotal data size = %g MB\n", total_data_size/(1024*1024.));



  //determine block size
  blocksize=VOLUME/numthreads;
  
  if(numthreads*blocksize<VOLUME){
    // do we have to re-size to get the extra??

    printf("resizing blocks %d -> ",blocksize);
    blocksize++;
    printf("%d\n",blocksize);
  }
  num_blocks = numthreads;
  printf("BLOCKSIZE\t%d\n",blocksize);

  //determine number of blocks  
  /*
  if(argc>1){
    blocksize =atoi(argv[1]);
    assert(VOLUME%blocksize == 0); // make life easy on ourselves
    printf("Using new blocksize: %d\n",blocksize);
  }
  num_blocks = VOLUME/blocksize;
  */
  printf("block data size\t%g MB\n",(float)blocksize*total_data_size/(VOLUME*(1024*1024.)));



  // INITIALIZATION OF VECTORS
  for(i=0; i< SU3VEC_SIZE; i++){
    // threading to speed up initialization
    // not part of the test
#pragma omp parallel for 
    for(j=0; j< VOLUME; j++){

      // initialize vecA with some stuff
      vecA.vec[i].re[j]=(float)(i+2);
      vecA.vec[i].im[j]=(float)(i-2);

      // initialize vecA with zeros
      vecB.vec[i].re[j]=0.0;
      vecB.vec[i].im[j]=0.0;
    }// volume loop
  }// i suvec loop


  // initialize with some sample SU(3) fields
  init_su3mats(&matM);


  // start timer
  tstart = dtime();

  // do some repeats so we do enough to time 
  for(rep=0; rep < num_reps; rep++){


 
#pragma omp parallel for private(indx, offset, lo, hi, i, j) schedule(static)
    // default is number of blocks is number of threads, unless user changes it
    for(tindex=0; tindex<num_blocks; tindex++){
      offset=tindex*blocksize;
      
      lo=offset;
      hi=offset+blocksize;
      if (hi > VOLUME){
	hi = VOLUME;
      }


      // each thread does this range of index values	 
      for(indx=lo; indx<hi ; indx++){
	// for the vecB assignments
	// (8 + 4 flops for each ) * 6   (3 components, 2 complex parts)
	// 72 flops per iteration
	//
	// with FORMAT_2COL, 
	// we have 7 additional flops for each assignment of
	//      m02r, m02i, m12r, m12i, m22r, m22i
	// so 42 more ops per iteration here 
	// 114 flops per iteration total


	// in each iteration
	// 3-col format 18+6 REALS in 6 REALS out = 30 REALS
	//              120 B for float 240 B for double
	/// 2-col format 12+6 REALS in 6 REALS out = 24
	//              96B for float, 192B for double 
#ifdef FORMAT_2COL  // do things differently for 2-column format
	REAL m02r;
	REAL m02i;

	REAL m12r;
	REAL m12i;

	REAL m22r;
	REAL m22i;



	m02r=
	  (
	   matM.mat[IDX_10].re[indx]*matM.mat[IDX_21].re[indx]
	   - matM.mat[IDX_10].im[indx]*matM.mat[IDX_21].im[indx]
	   
	   - matM.mat[IDX_11].re[indx]*matM.mat[IDX_20].re[indx]
	   + matM.mat[IDX_11].im[indx]*matM.mat[IDX_20].im[indx]
	   );
	
	m02i = 
	  -(
	    matM.mat[IDX_10].re[indx]*matM.mat[IDX_21].im[indx]
	    + matM.mat[IDX_10].im[indx]*matM.mat[IDX_21].re[indx]
	    
	    - matM.mat[IDX_11].re[indx]*matM.mat[IDX_20].im[indx]
	    - matM.mat[IDX_11].im[indx]*matM.mat[IDX_20].re[indx]
	    );
	
#endif	  


	    //#ops is 8 per iteration of inner loop
	    //16 reals = 12 in 4 out = 64B for floats
	    //ij=00
	    vecB.vec[0].re[indx] =  
	      + matM.mat[IDX_00].re[indx] * vecA.vec[0].re[indx]
	      - matM.mat[IDX_00].im[indx] * vecA.vec[0].im[indx]
	      //ij=01
	      + matM.mat[IDX_01].re[indx] * vecA.vec[1].re[indx]	    
	      -matM.mat[IDX_01].im[indx] * vecA.vec[1].im[indx]
#ifdef FORMAT_2COL
	      //ij=02
	      +m02r * vecA.vec[2].re[indx]
	      -m02i * vecA.vec[2].im[indx]
#else
	      +matM.mat[IDX_02].re[indx] * vecA.vec[2].re[indx]	    
	      -matM.mat[IDX_02].im[indx] * vecA.vec[2].im[indx]
#endif
	      ;


	    vecB.vec[0].im[indx] = 
	      //ij=00
	      - matM.mat[IDX_00].re[indx] * vecA.vec[0].im[indx]
	      - matM.mat[IDX_00].im[indx] * vecA.vec[0].re[indx]
	      //ij=01
	      -matM.mat[IDX_01].re[indx] * vecA.vec[1].im[indx]
	      -matM.mat[IDX_01].im[indx] * vecA.vec[1].re[indx]
#ifdef FORMAT_2COL
	      //ij=02
	      +m02r * vecA.vec[2].im[indx]
	      -m02i * vecA.vec[2].re[indx]
#else
	      -matM.mat[IDX_02].re[indx] * vecA.vec[2].im[indx]
	      -matM.mat[IDX_02].im[indx] * vecA.vec[2].re[indx]
#endif
	      ;



#ifdef FORMAT_2COL
	    m12r = 
	      (
	       matM.mat[IDX_01].re[indx]*matM.mat[IDX_20].re[indx]
	       - matM.mat[IDX_01].im[indx]*matM.mat[IDX_20].im[indx]
	       
	       - matM.mat[IDX_00].re[indx]*matM.mat[IDX_21].re[indx]
	       + matM.mat[IDX_00].im[indx]*matM.mat[IDX_21].im[indx]
	       );
	    
	    m12i = 
	      -(
		matM.mat[IDX_01].re[indx]*matM.mat[IDX_20].im[indx]
		+ matM.mat[IDX_01].im[indx]*matM.mat[IDX_20].re[indx]
		
		- matM.mat[IDX_00].re[indx]*matM.mat[IDX_21].im[indx]
		- matM.mat[IDX_00].im[indx]*matM.mat[IDX_21].re[indx]
		);
#endif

	    //ij=10
	    vecB.vec[1].re[indx] = 
	      +matM.mat[IDX_10].re[indx] * vecA.vec[0].re[indx]	    
	      -matM.mat[IDX_10].im[indx] * vecA.vec[0].im[indx]
	      //ij=11
	      +matM.mat[IDX_11].re[indx] * vecA.vec[1].re[indx]
	      -matM.mat[IDX_11].im[indx] * vecA.vec[1].im[indx]
	      //ij=12
#ifdef FORMAT_2COL
	      +m12r * vecA.vec[2].re[indx]
	      -m12i * vecA.vec[2].im[indx]
#else	     
	      +matM.mat[IDX_12].re[indx] * vecA.vec[2].re[indx]
	      -matM.mat[IDX_12].im[indx] * vecA.vec[2].im[indx]
#endif
	      ;
	    
	    //ij=10
	    vecB.vec[1].im[indx] = 
	      +matM.mat[IDX_10].re[indx] * vecA.vec[0].im[indx]
	      +matM.mat[IDX_10].im[indx] * vecA.vec[0].re[indx]
	      //ij=11
	      +matM.mat[IDX_11].re[indx] * vecA.vec[1].im[indx]
	      +matM.mat[IDX_11].im[indx] * vecA.vec[1].re[indx]
	      //ij=12
#ifdef FORMAT_2COL
	      +m12r * vecA.vec[2].im[indx]
	      -m12i * vecA.vec[2].re[indx]
#else
	      +matM.mat[IDX_12].re[indx] * vecA.vec[1].im[indx]
	      +matM.mat[IDX_12].im[indx] * vecA.vec[1].re[indx]
#endif
	      ;	      
	      
#ifdef FORMAT_2COL
	    m22r = 
	      (
	matM.mat[IDX_00].re[indx]*matM.mat[IDX_11].re[indx]
	  - matM.mat[IDX_00].im[indx]*matM.mat[IDX_11].im[indx]
	  
	  - matM.mat[IDX_01].re[indx]*matM.mat[IDX_10].re[indx]
	  + matM.mat[IDX_01].im[indx]*matM.mat[IDX_10].im[indx]
	  );
	    
	    m22i = 
	      -(
	+matM.mat[IDX_00].re[indx]*matM.mat[IDX_11].im[indx]
	  + matM.mat[IDX_00].im[indx]*matM.mat[IDX_11].re[indx]
	  
	  - matM.mat[IDX_01].re[indx]*matM.mat[IDX_10].im[indx]
	  - matM.mat[IDX_01].im[indx]*matM.mat[IDX_10].re[indx]
	  );
#endif

	    //ij=20
	    vecB.vec[2].re[indx] = 
	      +matM.mat[IDX_20].re[indx] * vecA.vec[0].re[indx]	    
	      -matM.mat[IDX_20].im[indx] * vecA.vec[0].im[indx]
	      //ij=21
	      +matM.mat[IDX_21].re[indx] * vecA.vec[1].re[indx]	    
	      -matM.mat[IDX_21].im[indx] * vecA.vec[1].im[indx]
	      //ij=22
#ifdef FORMAT_2COL
	      +m22r * vecA.vec[2].re[indx]
	      -m22i * vecA.vec[2].im[indx]
#else
	      +matM.mat[IDX_22].re[indx] * vecA.vec[2].re[indx]
	      -matM.mat[IDX_22].im[indx] * vecA.vec[2].im[indx]
#endif
	      ;
	    
	    vecB.vec[2].im[indx] = 
	      //ij=20
	      +matM.mat[IDX_20].re[indx] * vecA.vec[0].im[indx]
	      +matM.mat[IDX_20].im[indx] * vecA.vec[0].re[indx]
	      //ij=21
	      +matM.mat[IDX_21].re[indx] * vecA.vec[1].im[indx]
	      +matM.mat[IDX_21].im[indx] * vecA.vec[1].re[indx]
	      //ij=22
#ifdef FORMAT_2COL
	      +m22r * vecA.vec[2].im[indx]
	      -m22i * vecA.vec[2].re[indx]
#else
	      +matM.mat[IDX_22].re[indx] * vecA.vec[2].im[indx]
	      +matM.mat[IDX_22].im[indx] * vecA.vec[2].re[indx]
#endif
	      ;

	    

      }//indx

    }//tindex
    
  }//reps

    tstop = dtime();
  //elasped time
  ttime = tstop - tstart;
  //
  
  ops_per_iter_1 = 72;
  bytes_per_iter_1 = 16*sizeof(REAL)*SU3MAT_SIZE;

  if(SU3MAT_SIZE==9){
    // 3-col format
    ops_per_iter_2=0;
    bytes_per_iter_2 = 30*sizeof(REAL);
  }else{
    //2-col format
    ops_per_iter_2 = 42;
    bytes_per_iter_2 = 24*sizeof(REAL);
  }

  // # of gigaflops we just calculated  
  gflops = (double)( 1.0e-9*num_reps*VOLUME*(ops_per_iter_1 
					     + ops_per_iter_2));
  //  gbytes = (double)( 1.0e-9*num_reps*total_data_size);
  gbytes = (double)( 1.0e-9*num_reps*VOLUME*bytes_per_iter_2);

  gflop_per_sec= gflops/ttime ;
  bandwidth = gbytes/ttime ;



  printf("Using %s format\n",FORMAT_LABEL);
  printf(" threads = %d\n",numthreads);
  printf("total gflops: %f\n", gflops);
  printf("total gbytes: %f\n", gbytes);
  printf ("elapsed time (s) = %f\n",ttime);
  printf("PERF:\tGflop/sec = %g\n",gflop_per_sec);
  printf("Bandwidth GB/sec = %g\n",bandwidth);

  /*
  printf("op/iter = %d\n",ops_per_iter);
  printf("byte/iter = %d\n",bytes_per_iter);
  */

  printf("Intensity (flops/byte)= %f\n", gflops/gbytes);

}
